/******************************************************************
Copyright 2000 by Object Craft P/L, Melbourne, Australia.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft
is not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

OBJECT CRAFT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#include "mssqldb.h"

int is_fixedlen(int vartype)
{
    if (vartype == SQLCHAR || vartype == SQLBINARY)
	return 0;
    return 1;
}

PyObject *DataBuf_convert(DBPROCObj *dbproc, PyObject *from, DataBufObj *to)
{
    DBPROCESS *db_dbproc;
    RETCODE status;
    int fromtype, intbuff;
    double doublebuff;
    BYTE *frombuff;
    DBINT fromlen, tolen;

    if (dbproc != NULL)
	db_dbproc = dbproc->dbproc;
    else
	db_dbproc = NULL;
    if (DataBuf_Check(from)) {
	fromtype = ((DataBufObj*)from)->vartype;
	frombuff = ((DataBufObj*)from)->buff;
	if (is_fixedlen(fromtype))
	    fromlen = -1;
	else
	    fromlen = ((DataBufObj*)from)->varlen;
    } else if (PyString_Check(from)) {
	fromtype = SQLCHAR;
	frombuff = PyString_AsString(from);
	fromlen = PyString_Size(from);
    } else if (PyInt_Check(from)) {
	fromtype = SQLINT4;
	intbuff = PyInt_AsLong(from);
	frombuff = (BYTE*)&intbuff;
	fromlen = -1;
    } else if (!PyFloat_Check(from)) {
	fromtype = SQLFLT8;
	doublebuff = PyFloat_AsDouble(from);
	frombuff = (BYTE*)&doublebuff;
	fromlen = -1;
    } else if (!Numeric_Check(from)) {
	fromtype = SQLNUMERIC;
	frombuff = (BYTE*)&((NumericObj*)from)->num;
	fromlen = -1;
    } else if (!DateTime_Check(from)) {
	fromtype = SQLDATETIME;
	frombuff = (BYTE*)&((DateTimeObj*)from)->datetime;
	fromlen = -1;
    } else {
	PyErr_SetString(PyExc_TypeError, "unrecognised from data");
	return NULL;
    }
    if (is_fixedlen(to->vartype))
	tolen = -1;
    else
	tolen = to->size;

    PyErr_Clear();
    status = dbconvert(db_dbproc,
		       fromtype, frombuff, fromlen,
		       to->vartype, to->buff, tolen);
    if (PyErr_Occurred())
	return NULL;

    if (status != -1 && status != FAIL)
	to->varlen = status;
    return PyInt_FromLong(status);
}

static PyObject *get_buff_value(DataBufObj *self)
{
    switch (self->vartype) {
    case SQLBINARY:
    case SQLCHAR:
	return PyString_FromStringAndSize(((DBVARYCHAR*)self->buff)->str,
					  ((DBVARYCHAR*)self->buff)->len);

    case SQLBIT:
	return PyInt_FromLong(*(DBBIT*)self->buff);

    case SQLINT1:
	return PyInt_FromLong(*(DBTINYINT*)self->buff);

    case SQLINT2:
	return PyInt_FromLong(*(DBSMALLINT*)self->buff);

    case SQLINT4:
	return PyInt_FromLong(*(DBINT*)self->buff);

    case SQLMONEY:
	return money_alloc((DBMONEY*)self->buff);

    case SQLFLT4:
	return PyFloat_FromDouble(*(DBFLT4*)self->buff);

    case SQLFLT8:
	return PyFloat_FromDouble(*(DBFLT8*)self->buff);

    case SQLDATETIME:
	return datetime_alloc((DBDATETIME*)self->buff);

    case SQLNUMERIC:
    case SQLDECIMAL:
	return numeric_alloc((DBNUMERIC*)self->buff);

    default:
	PyErr_SetString(PyExc_TypeError, "unknown data format");
	return NULL;
    }
}

static char DataBuf_get__doc__[] = 
"get() -> value";

static PyObject *DataBuf_get(DataBufObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    return get_buff_value(self);
}

static char DataBuf_set__doc__[] = 
"set(value) -> None";

static PyObject *DataBuf_set(DataBufObj *self, PyObject *args)
{
    PyObject *obj, *tmp_obj = NULL;
    int size;

    if (!PyArg_ParseTuple(args, "O", &obj))
	return NULL;

    switch (self->vartype) {
    case SQLBINARY:
    case SQLCHAR:
	if (!PyString_Check(obj)) {
	    tmp_obj = PyObject_Str(obj);
	    if (tmp_obj == NULL)
		return NULL;
	    obj = tmp_obj;
	}
	size = PyString_Size(obj);
	if (size > sizeof(((DBVARYCHAR*)self->buff)->str)) {
	    PyErr_SetString(PyExc_TypeError, "string too long for buffer");
	    Py_XDECREF(tmp_obj);
	    return NULL;
	}
	memmove(((DBVARYCHAR*)self->buff)->str, PyString_AsString(obj), size);
	((DBVARYCHAR*)self->buff)->len = size;
	break;

    case SQLBIT:
	if (!PyInt_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "integer expected");
	    return NULL;
	}
	*(DBBIT*)self->buff = (DBBIT)PyInt_AsLong(obj);
	self->varlen = self->size;
	break;

    case SQLINT1:
	if (!PyInt_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "integer expected");
	    return NULL;
	}
	*(DBTINYINT*)self->buff = (DBTINYINT)PyInt_AsLong(obj);
	self->varlen = self->size;
	break;

    case SQLINT2:
	if (!PyInt_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "integer expected");
	    return NULL;
	}
	*(DBSMALLINT*)self->buff = (DBSMALLINT)PyInt_AsLong(obj);
	self->varlen = self->size;
	break;

    case SQLINT4:
	if (!PyInt_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "integer expected");
	    return NULL;
	}
	*(DBINT*)self->buff = PyInt_AsLong(obj);
	self->varlen = self->size;
	break;

    case SQLMONEY:
	if (!Money_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "money expected");
	    return NULL;
	}
	*(DBMONEY*)self->buff = ((MoneyObj*)obj)->money;
	self->varlen = self->size;
	break;

    case SQLFLT8:
	if (!PyFloat_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "float expected");
	    return NULL;
	}
	*(DBFLT8*)self->buff = (DBFLT8)PyFloat_AsDouble(obj);
	self->varlen = self->size;
	break;

    case SQLFLT4:
	if (!PyFloat_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "float expected");
	    return NULL;
	}
	*(DBFLT4*)self->buff = (DBFLT4)PyFloat_AsDouble(obj);
	self->varlen = self->size;
	break;

    case SQLDATETIME:
	if (!DateTime_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "datetime expected");
	    return NULL;
	}
	*(DBDATETIME*)self->buff = ((DateTimeObj*)obj)->datetime;
	self->varlen = self->size;
	break;

    case SQLNUMERIC:
    case SQLDECIMAL:
	if (!Numeric_Check(obj)) {
	    PyErr_SetString(PyExc_TypeError, "numeric expected");
	    return NULL;
	}
	*(DBNUMERIC*)self->buff = ((NumericObj*)obj)->num;
	self->varlen = self->size;
	break;

    default:
	PyErr_SetString(PyExc_TypeError, "unknown data format");
	return NULL;
    }

    Py_XDECREF(tmp_obj);
    Py_INCREF(Py_None);
    return Py_None;
}

static struct PyMethodDef DataBuf_methods[] = {
    { "get", (PyCFunction)DataBuf_get, METH_VARARGS, DataBuf_get__doc__ },
    { "set", (PyCFunction)DataBuf_set, METH_VARARGS, DataBuf_set__doc__ },
    { NULL }
};

PyObject *DataBuf__init__(int vartype, int varlen)
{
    DataBufObj *self;

    switch (vartype) {
    case SQLBINARY:
	break;
    case SQLCHAR:
	varlen = sizeof(DBVARYCHAR);
	break;
    case SQLBIT:
	varlen = sizeof(DBBIT);
	break;
    case SQLINT1:
	varlen = sizeof(DBTINYINT);
	break;
    case SQLINT2:
	varlen = sizeof(DBSMALLINT);
	break;
    case SQLINT4:
	varlen = sizeof(DBINT);
	break;
    case SQLMONEY4:
	vartype = SQLMONEY;
	/* promote */
    case SQLMONEY:
	varlen = sizeof(DBMONEY);
	break;
    case SQLFLT4:
	varlen = sizeof(DBFLT4);
	break;
    case SQLFLT8:
	varlen = sizeof(DBFLT8);
	break;
    case SQLDATETIM4:
	vartype = SQLDATETIME;
	/* promote */
    case SQLDATETIME:
	varlen = sizeof(DBDATETIME);
	break;
    case SQLNUMERIC:
    case SQLDECIMAL:
	varlen = sizeof(DBNUMERIC);
	break;

    default:
	PyErr_SetString(PyExc_TypeError, "unknown data format");
	return NULL;
    }

    self = PyObject_NEW(DataBufObj, &DataBufType);
    if (self == NULL)
	return NULL;

    self->strip = 0;
    self->vartype = vartype;
    self->size = self->varlen = varlen;
    self->buff = malloc(self->size);
    if (self->buff == NULL) {
	Py_DECREF(self);
	return NULL;
    }

    return (PyObject*)self;
}

static void DataBuf__del__(DataBufObj *self)
{
    if (self->buff != NULL)
	free(self->buff);

    PyMem_DEL(self);
}

#define OFFSET(x) offsetof(DataBufObj, x)

static struct memberlist DataBuf_memberlist[] = {
    { "type", T_INT, OFFSET(vartype) },
    { "len",  T_INT, OFFSET(varlen) },
    { "size", T_INT, OFFSET(size), RO },
    { NULL }
};

static PyObject *DataBuf_getattr(DataBufObj *self, char *name)
{
    PyObject *rv;

    rv = PyMember_Get((char *)self, DataBuf_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(DataBuf_methods, (PyObject *)self, name);
}

static int DataBuf_setattr(DataBufObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "cannot delete attribute");
	return -1;
    }
    return PyMember_Set((char *)self, DataBuf_memberlist, name, v);
}

PyObject *DataBuf__str__(DataBufObj *self)
{
    PyObject *obj;
    PyObject *str_value;

    obj = get_buff_value(self);
    if (obj == NULL)
	return NULL;
    str_value = PyObject_Str(obj);
    Py_DECREF(obj);
    return str_value;
}

static char DataBufType__doc__[] = 
"";

PyTypeObject DataBufType = {
    PyObject_HEAD_INIT(0)
    0,				/*ob_size*/
    "DataBufType",		/*tp_name*/
    sizeof(DataBufObj),		/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)DataBuf__del__,	/*tp_dealloc*/
    (printfunc)0,		/*tp_print*/
    (getattrfunc)DataBuf_getattr, /*tp_getattr*/
    (setattrfunc)DataBuf_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)0,		/*tp_repr*/
    0,				/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)DataBuf__str__,	/*tp_str*/

    0L, 0L, 0L, 0L,
    DataBufType__doc__
};
